/*
 * Copyright (c) 2001 Mark Fullmer and The Ohio State University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      $Id: flow-export.c,v 1.5 2001/02/24 21:35:45 maf Exp $
 */

#if HAVE_CONFIG_H
 #include <config.h>
#endif

#include <sys/time.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#if HAVE_STRINGS_H
 #include <strings.h>
#endif
#if HAVE_STRING_H
  #include <string.h>
#endif
#include <time.h>
#include <fcntl.h>
#include <ftlib.h>
#include <sym.h>
#include "pcap.h"
#include "cflowd.h"

struct options {
  u_int32 cflowd_mask;
};

struct jump {
    int (*where)(struct ftio *ftio, struct options *opt);
};

int format0(struct ftio *ftio, struct options *opt);
int format1(struct ftio *ftio, struct options *opt);

void usage(void);

struct jump format[] = {{format0}, {format1}};

#define NFORMATS 2 /* nformats + 1 */

int main(int argc, char **argv)
{
  int i, format_index, ret;
  extern char *optarg;
  extern struct jump format[];
  struct ftio ftio;
  struct ftprof ftp;
  struct options opt;
  int debug;

  debug = 0;
  format_index = 0;
  bzero(&opt, sizeof opt);

  opt.cflowd_mask = CF_ROUTERMASK | CF_SRCIPADDRMASK | CF_DSTIPADDRMASK |
    CF_INPUTIFINDEXMASK | CF_OUTPUTIFINDEXMASK | CF_SRCPORTMASK |
    CF_DSTPORTMASK | CF_PKTSMASK | CF_BYTESMASK | CF_IPNEXTHOPMASK |
    CF_STARTTIMEMASK | CF_ENDTIMEMASK | CF_PROTOCOLMASK | CF_TOSMASK |
    CF_SRCASMASK | CF_DSTASMASK | CF_SRCMASKLENMASK | CF_DSTMASKLENMASK |
    CF_TCPFLAGSMASK | CF_INPUTENCAPMASK | CF_OUTPUTENCAPMASK |
    CF_PEERNEXTHOPMASK | CF_ENGINETYPEMASK | CF_ENGINEIDMASK;

  /* profile */
  ftprof_start (&ftp);

  while ((i = getopt(argc, argv, "h?d:f:m:")) != -1)
    switch (i) {

    case 'd': /* debug */
      debug = atoi(optarg);
      break;

    case 'f': /* format */
      format_index = atoi(optarg);
      break;

    case 'h': /* help */
    case '?':
      usage();
      exit (0);
      break;

    case 'm': /* cflowd mask */
      opt.cflowd_mask = strtoul(optarg, (char **)0L, 10);

    default:
      usage();
      exit (1);
      break;

    } /* switch */

  if (format_index >= NFORMATS) {
    fprintf(stderr, "no such format, %d\n", format_index);
    exit (1);
  }

  /* read from stdin */
  if (ftio_init(&ftio, 0, FT_IO_FLAG_READ) < 0) {
    fprintf(stderr, "ftio_init(): failed\n");
    exit (1);
  }

  ret = format[format_index].where(&ftio, &opt);
      
  if ((!ret) && (debug > 0)) {
    ftprof_end(&ftp, ftio_get_rec_total(&ftio));
    ftprof_print(&ftp, argv[0], stderr);
  }
        
  return ret;

} /* main */

void usage() {

  fprintf(stderr, "flow-export:\n\n");
  fprintf(stderr, " -d # set debug level.\n");
  fprintf(stderr, " -f # format\n");
  fprintf(stderr, "    0 cflowd\n");
  fprintf(stderr, "    1 pcap (tcpdump)\n");
  fprintf(stderr, " -m # 32bit mask for cflowd export\n");
  fprintf(stderr, "\n\n");

} /* usage */

/*
 * function: format0
 *
 * export flows in cflowd format
*/
int format0(struct ftio *ftio, struct options *opt)
{
  struct fts3rec_gen *rec;
  char buf[1024];
  int ret, bsize, bsize2;
  u_int32 ui32;
  u_int16 ui16;

  /* stash away the router IP address in network byte order */

  /* foreach flow */
    /* convert record to network byte order */
    /* copy each individual record to buf */
    /* fwrite buf */
  /* done */

  return 0;
} /* format 0 */

/*
 * function: format1
 *
 * export flows in pcap format.  Hack to use tcpdump's packet matcher
*/
int format1(struct ftio *ftio, struct options *opt)
{
  struct fts3rec_gen *rec;
  struct pcap_file_header pfh;
  struct pcap_packet_header pph;
  struct pcap_data1 pd1;
  struct pcap_data2 pd2;
  struct pcap_data3 pd3;
  struct pcap_data4 pd4;
  int good;
  struct timeval now;
  struct timezone tz;
  long thiszone;
  char buf[1024];
  int bsize, bsize2;

  if (ftio_check_generic(ftio) < 0)
    return -1;

  if (gettimeofday(&now, &tz) < 0) {
    fprintf(stderr, "gettimeofday() failed\n");
    return -1;
  }

  bzero(&pfh, sizeof pfh);
  bzero(&pph, sizeof pph);
  bzero(&pd1, sizeof pd1);
  bzero(&pd2, sizeof pd2);
  bzero(&pd3, sizeof pd3);
  bzero(&pd4, sizeof pd4);
  bsize = 0;

  thiszone = tz.tz_minuteswest * -60;

  if (localtime((time_t *)&now.tv_sec)->tm_isdst)
    thiszone += 3600;

  pfh.magic = TCPDUMP_MAGIC;  
  pfh.version_major = TCPDUMP_VERSION_MAJOR;
  pfh.version_minor = TCPDUMP_VERSION_MINOR;
  pfh.thiszone = thiszone;
  pfh.sigfigs = 6;
  pfh.snaplen = 38; /* XXX TODO */
  pfh.linktype = 1;

  if (fwrite(&pfh, sizeof pfh, 1, stdout) != 1) {
    fprintf(stderr, "pcap header write failed.\n");
    return -1;
  }

  pph.len = 58;
  pph.caplen = 58;

  pd1.eth_prot = 0x0008;
  pd2.version = 0x45;

  bcopy(&pph, buf, sizeof pph);
  bsize += sizeof pph;

  bcopy(&pd1, buf+bsize, sizeof pd1);
  bsize += sizeof pd1;

  while ((rec = ftio_read(ftio))) {

    pd2.tos = rec->tos;
    pd2.prot = rec->prot;
    pd2.srcaddr = rec->srcaddr;
    pd2.dstaddr = rec->dstaddr;

#if BYTE_ORDER == LITTLE_ENDIAN
    SWAPINT32(pd2.srcaddr);
    SWAPINT32(pd2.dstaddr);
#endif /* LITTLE_ENDIAN */

    good = 1;

    switch (pd2.prot) {

    case 6:

      pd3.srcport = rec->srcport;
      pd3.dstport = rec->dstport;

#if BYTE_ORDER == LITTLE_ENDIAN
      SWAPINT16(pd3.srcport);
      SWAPINT16(pd3.dstport);
#endif /* LITTLE_ENDIAN */

      bcopy(&pd2, buf+bsize, sizeof pd2);
      bcopy(&pd3, buf+bsize+sizeof pd2, sizeof pd3);
      bsize2 = bsize + sizeof pd2 + sizeof pd3;

      break;

    case 17:

      pd4.srcport = rec->srcport;
      pd4.dstport = rec->dstport;

#if BYTE_ORDER == LITTLE_ENDIAN
      SWAPINT16(pd4.srcport);
      SWAPINT16(pd4.dstport);
#endif /* LITTLE_ENDIAN */

      bcopy(&pd2, buf+bsize, sizeof pd2);
      bcopy(&pd4, buf+bsize+sizeof pd2, sizeof pd4);
      bsize2 = bsize + sizeof pd2 + sizeof pd4;

      break;

    default:
      good = 0;
      break;

    } /* switch */

    if (good) {
      if (fwrite(&buf, bsize2, 1, stdout) != 1) {
        fprintf(stderr, "pcap pkt write failed.\n");
        return -1;
      }
    }

  } /* while */

  return 0;
  
} /* format1 */


